@@ -38,7 +38,6 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	WacomDevicePtr   priv   = NULL;
 	WacomCommonPtr   common = NULL;
 	WacomToolPtr     tool   = NULL;
-	WacomToolAreaPtr area   = NULL;
 	int i;
 
 	priv = calloc(1, sizeof(WacomDeviceRec));
@@ -53,10 +52,6 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	if(!tool)
 		goto error;
 
-	area = calloc(1, sizeof(WacomToolArea));
-	if (!area)
-		goto error;
-
 	pInfo->device_control = gWacomModule.DevProc;
 	pInfo->read_input = gWacomModule.DevReadInput;
 	pInfo->control_proc = gWacomModule.DevChangeControl;
@@ -67,7 +62,7 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	priv->next = NULL;
 	priv->pInfo = pInfo;
 	priv->common = common;       /* common info pointer */
-	priv->oldHwProx = 1;	     /* previous hardware proximity */
+	priv->oldCursorHwProx = 0;   /* previous cursor hardware proximity */
 	priv->nPressCtrl [0] = 0;    /* pressure curve x0 */
 	priv->nPressCtrl [1] = 0;    /* pressure curve y0 */
 	priv->nPressCtrl [2] = 100;  /* pressure curve x1 */
@@ -76,70 +71,42 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	/* Default button and expresskey values, offset buttons 4 and higher
 	 * by the 4 scroll buttons. */
 	for (i=0; i<WCM_MAX_BUTTONS; i++)
-		priv->button[i] = (i < 3) ? i + 1 : i + 5;
+		priv->button_default[i] = (i < 3) ? i + 1 : i + 5;
 
-	priv->nbuttons = WCM_MAX_BUTTONS;		/* Default number of buttons */
-	priv->relup = 5;			/* Default relative wheel up event */
-	priv->reldn = 4;			/* Default relative wheel down event */
+	priv->nbuttons = WCM_MAX_BUTTONS;       /* Default number of buttons */
+	priv->wheel_default[WHEEL_REL_UP] = 5;
+	priv->wheel_default[WHEEL_REL_DN] = 4;
 	/* wheel events are set to 0, but the pad overwrites this default
 	 * later in wcmParseOptions, when we have IsPad() available */
-	priv->wheelup = 0;			/* Default absolute wheel up event */
-	priv->wheeldn = 0;			/* Default absolute wheel down event */
-	priv->striplup = 4;			/* Default left strip up event */
-	priv->stripldn = 5;			/* Default left strip down event */
-	priv->striprup = 4;			/* Default right strip up event */
-	priv->striprdn = 5;			/* Default right strip down event */
+	priv->wheel_default[WHEEL_ABS_UP] = 0;
+	priv->wheel_default[WHEEL_ABS_DN] = 0;
+	priv->wheel_default[WHEEL2_ABS_UP] = 0;
+	priv->wheel_default[WHEEL2_ABS_DN] = 0;
+	priv->strip_default[STRIP_LEFT_UP] = 4;
+	priv->strip_default[STRIP_LEFT_DN] = 5;
+	priv->strip_default[STRIP_RIGHT_UP] = 4;
+	priv->strip_default[STRIP_RIGHT_DN] = 5;
 	priv->naxes = 6;			/* Default number of axes */
 
 	/* JEJ - throttle sampling code */
 	priv->throttleLimit = -1;
 
-	common->wcmFlags = RAW_FILTERING_FLAG;   /* various flags */
 	common->wcmDevices = priv;
-	common->wcmProtocolLevel = WCM_PROTOCOL_4; /* protocol level */
-	common->wcmTPCButton = 0;          /* set Tablet PC button on/off */
-	common->wcmCapacity = -1;          /* Capacity is disabled */
-	common->wcmCapacityDefault = -1;    /* default to -1 when capacity isn't supported */
-					   /* 3 when capacity is supported */
-	common->wcmGestureParameters.wcmZoomDistance = 50;
-	common->wcmGestureParameters.wcmZoomDistanceDefault = 50;
-	common->wcmGestureParameters.wcmScrollDirection = 0;
-	common->wcmGestureParameters.wcmScrollDistance = 20;
-	common->wcmGestureParameters.wcmScrollDistanceDefault = 20;
-	common->wcmGestureParameters.wcmTapTime = 250;
-	common->wcmGestureParameters.wcmTapTimeDefault = 250;
-	common->wcmRotate = ROTATE_NONE;   /* default tablet rotation to off */
-	common->wcmMaxX = 0;               /* max digitizer logical X value */
-	common->wcmMaxY = 0;               /* max digitizer logical Y value */
-	common->wcmMaxTouchX = 1024;       /* max touch X value */
-	common->wcmMaxTouchY = 1024;       /* max touch Y value */
-	common->wcmMaxStripX = 4096;       /* Max fingerstrip X */
-	common->wcmMaxStripY = 4096;       /* Max fingerstrip Y */
-	common->wcmMaxtiltX = 128;	   /* Max tilt in X directory */
-	common->wcmMaxtiltY = 128;	   /* Max tilt in Y directory */
-	common->wcmCursorProxoutDistDefault = PROXOUT_INTUOS_DISTANCE; 
-			/* default to Intuos */
-	common->wcmSuppress = DEFAULT_SUPPRESS;    
-			/* transmit position if increment is superior */
-	common->wcmRawSample = DEFAULT_SAMPLES;    
-			/* number of raw data to be used to for filtering */
 
 	/* tool */
 	priv->tool = tool;
 	common->wcmTool = tool;
 	tool->next = NULL;          /* next tool in list */
-	tool->arealist = area;      /* list of defined areas */
+	tool->device = pInfo;
 	/* tool->typeid is set once we know the type - see wcmSetType */
 
-	/* tool area */
-	priv->toolarea = area;
-	area->next = NULL;    /* next area in list */
-	area->device = pInfo; /* associated WacomDevice */
+	/* timers */
+	priv->serial_timer = TimerSet(NULL, 0, 0, NULL, NULL);
+	priv->tap_timer = TimerSet(NULL, 0, 0, NULL, NULL);
 
 	return 1;
 
 error:
-	free(area);
 	free(tool);
 	wcmFreeCommon(&common);
 	free(priv);
@@ -158,7 +125,8 @@ static void wcmFree(InputInfoPtr pInfo)
 	if (!priv)
 		return;
 
-	free(priv->toolarea);
+	TimerFree(priv->serial_timer);
+	TimerFree(priv->tap_timer);
 	free(priv->tool);
 	wcmFreeCommon(&priv->common);
 	free(priv);
@@ -171,12 +139,7 @@ static int wcmSetType(InputInfoPtr pInfo, const char *type)
 	WacomDevicePtr priv = pInfo->private;
 
 	if (!type)
-	{
-		xf86Msg(X_ERROR, "%s: No type or invalid type specified.\n"
-				"Must be one of stylus, touch, cursor, eraser, or pad\n",
-				pInfo->name);
-		return 0;
-	}
+		goto invalid;
 
 	if (xf86NameCmp(type, "stylus") == 0)
 	{
@@ -203,7 +166,8 @@ static int wcmSetType(InputInfoPtr pInfo, const char *type)
 	{
 		priv->flags = ABSOLUTE_FLAG|PAD_ID;
 		pInfo->type_name = WACOM_PROP_XI_TYPE_PAD;
-	}
+	} else
+		goto invalid;
 
 	/* Set the device id of the "last seen" device on this tool */
 	priv->old_device_id = wcmGetPhyDeviceID(priv);
@@ -214,6 +178,12 @@ static int wcmSetType(InputInfoPtr pInfo, const char *type)
 	priv->tool->typeid = DEVICE_ID(priv->flags); /* tool type (stylus/touch/eraser/cursor/pad) */
 
 	return 1;
+
+invalid:
+	xf86Msg(X_ERROR, "%s: No type or invalid type specified.\n"
+			 "Must be one of stylus, touch, cursor, eraser, or pad\n",
+			 pInfo->name);
+	return 0;
 }
 
 int wcmGetPhyDeviceID(WacomDevicePtr priv)
@@ -239,7 +209,7 @@ int wcmGetPhyDeviceID(WacomDevicePtr priv)
  * starts making sense again.
  */
 
-static char *default_options[] =
+static const char *default_options[] =
 {
 	"StopBits",    "1",
 	"DataBits",    "8",
@@ -257,10 +227,17 @@ static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	WacomDevicePtr dev;
 	WacomDevicePtr *prev;
-	WacomCommonPtr common = priv->common;
+	WacomCommonPtr common;
+
+	if (!priv)
+		goto out;
+
+	common = priv->common;
 
 	DBG(1, priv, "\n");
 
+	/* Server 1.10 will UnInit all devices for us */
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 	if (priv->isParent)
 	{
 		/* HAL removal sees the parent device removed first. */
@@ -285,22 +262,7 @@ static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		free(pInfo->name);
 		pInfo->name = NULL;
 	}
-
-	if (priv->toolarea)
-	{
-		WacomToolAreaPtr *prev_area = &priv->tool->arealist;
-		WacomToolAreaPtr area = *prev_area;
-		while (area)
-		{
-			if (area == priv->toolarea)
-			{
-				*prev_area = area->next;
-				break;
-			}
-			prev_area = &area->next;
-			area = area->next;
-		}
-	}
+#endif
 
 	if (priv->tool)
 	{
@@ -331,6 +293,7 @@ static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		dev = dev->next;
 	}
 
+out:
 	wcmFree(pInfo);
 	xf86DeleteInput(pInfo, 0);
 }
@@ -365,6 +328,11 @@ static Bool wcmMatchDevice(InputInfoPtr pLocal, WacomCommonPtr *common_return)
 		{
 			DBG(2, priv, "port share between %s and %s\n",
 					pLocal->name, pMatch->name);
+			/* FIXME: we loose the common->wcmTool here but it
+			 * gets re-added during wcmParseOptions. This is
+			 * currently required by the code, adding the tool
+			 * again here means we trigger the duplicate tool
+			 * detection */
 			wcmFreeCommon(&priv->common);
 			priv->common = wcmRefCommon(privMatch->common);
 			priv->next = priv->common->wcmDevices;
@@ -417,6 +385,66 @@ wcmInitModel(InputInfoPtr pInfo)
 	return TRUE;
 }
 
+/**
+ * Link the touch tool to the pen of the same device
+ * so we can arbitrate the events when posting them.
+ */
+static void wcmLinkTouchAndPen(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = pInfo->private;
+	WacomCommonPtr common = priv->common;
+	InputInfoPtr device = xf86FirstLocalDevice();
+	WacomCommonPtr tmpcommon = NULL;
+	WacomDevicePtr tmppriv = NULL;
+	Bool touch_device_assigned = FALSE;
+
+	/* Lookup to find the associated pen and touch */
+	for (; device != NULL; device = device->next)
+	{
+		if (!strcmp(device->drv->driverName, "wacom"))
+		{
+			tmppriv = (WacomDevicePtr) device->private;
+			tmpcommon = tmppriv->common;
+			touch_device_assigned = (common->wcmTouchDevice ||
+						tmpcommon->wcmTouchDevice);
+
+			/* skip the same tool or already linked devices */
+			if ((tmppriv == priv) || touch_device_assigned)
+				continue;
+
+			if (tmpcommon->tablet_id == common->tablet_id)
+			{
+				if (IsTouch(tmppriv) && IsPen(priv))
+					common->wcmTouchDevice = tmppriv;
+				else if (IsTouch(priv) && IsPen(tmppriv))
+					tmpcommon->wcmTouchDevice = priv;
+
+				if (common->wcmTouchDevice ||
+						tmpcommon->wcmTouchDevice)
+				{
+					TabletSetFeature(common, WCM_PENTOUCH);
+					TabletSetFeature(tmpcommon, WCM_PENTOUCH);
+				}
+			}
+		}
+	}
+}
+
+/**
+ * Check if this device was hotplugged by the driver by checking the _source
+ * option.
+ *
+ * Must be called before wcmNeedAutoHotplug()
+ *
+ * @return True if the source for this device is the wacom driver itself or
+ * false otherwise.
+ */
+static int wcmIsHotpluggedDevice(InputInfoPtr pInfo)
+{
+	char *source = xf86CheckStrOption(pInfo->options, "_source", "");
+	return !strcmp(source, "_driver/wacom");
+}
+
 /* wcmPreInit - called for each input devices with the driver set to
  * "wacom" */
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
@@ -455,13 +483,13 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	WacomDevicePtr priv = NULL;
 	WacomCommonPtr common = NULL;
 	const char*	type;
-	char*		device, *oldname;
-	int		need_hotplug = 0;
+	const char*	device, *oldname;
+	int		need_hotplug = 0, is_dependent = 0;
 
 	gWacomModule.wcmDrv = drv;
 
 	device = xf86SetStrOption(pInfo->options, "Device", NULL);
-	type = xf86FindOptionValue(pInfo->options, "Type");
+	type = xf86SetStrOption(pInfo->options, "Type", NULL);
 
 	/*
 	   Init process:
@@ -484,6 +512,8 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	priv = (WacomDevicePtr) pInfo->private;
 	priv->common->device_path = device;
 	priv->name = pInfo->name;
+	priv->debugLevel = xf86SetIntOption(pInfo->options,
+					    "DebugLevel", priv->debugLevel);
 
 	/* check if the same device file has been added already */
 	if (wcmIsDuplicate(device, pInfo))
@@ -501,9 +531,13 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		/* initialize supported keys with the first tool on the port */
 		wcmDeviceTypeKeys(pInfo);
 
+	common->debugLevel = xf86SetIntOption(pInfo->options,
+					      "CommonDBG", common->debugLevel);
 	oldname = pInfo->name;
 
-	if ((need_hotplug = wcmNeedAutoHotplug(pInfo, &type)))
+	if (wcmIsHotpluggedDevice(pInfo))
+		is_dependent = 1;
+	else if ((need_hotplug = wcmNeedAutoHotplug(pInfo, &type)))
 	{
 		/* we need subdevices, change the name so all of them have a
 		   type. */
@@ -520,12 +554,15 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	if (!wcmSetType(pInfo, type))
 		goto SetupProc_fail;
 
-	if (!wcmParseOptions(pInfo, need_hotplug))
+	if (!wcmPreInitParseOptions(pInfo, need_hotplug, is_dependent))
 		goto SetupProc_fail;
 
 	if (!wcmInitModel(pInfo))
 		goto SetupProc_fail;
 
+	if (!wcmPostInitParseOptions(pInfo, need_hotplug, is_dependent))
+		goto SetupProc_fail;
+
 	if (need_hotplug)
 	{
 		priv->isParent = 1;
@@ -538,6 +575,12 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		pInfo->fd = -1;
 	}
 
+	/* only link them once per port. We need to try for both pen and touch
+	 * since we do not know which tool (touch or pen) will be added first.
+	 */
+	if (IsTouch(priv) || (IsPen(priv) && !common->wcmTouchDevice))
+		wcmLinkTouchAndPen(pInfo);
+
 	return Success;
 
 SetupProc_fail:
@@ -545,7 +588,7 @@ SetupProc_fail:
 	if (common && priv)
 		common->wcmDevices = priv->next;
 
-	if (pInfo && pInfo->fd != -1)
+	if (pInfo->fd != -1)
 	{
 		close(pInfo->fd);
 		pInfo->fd = -1;
