@@ -33,15 +33,11 @@
 static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
 {
 	int match = 0;
-	char* device;
-	char* fsource = xf86CheckStrOption(pInfo->options, "_source", "");
 	InputInfoPtr pDevices = xf86FirstLocalDevice();
-	WacomCommonPtr pCommon = NULL;
-	char* psource;
 
 	for (; pDevices != NULL; pDevices = pDevices->next)
 	{
-		device = xf86CheckStrOption(pDevices->options, "Device", NULL);
+		char* device = xf86CheckStrOption(pDevices->options, "Device", NULL);
 
 		/* device can be NULL on some distros */
 		if (!device || !strstr(pDevices->drv->driverName, "wacom"))
@@ -49,8 +45,10 @@ static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
 
 		if (pInfo != pDevices)
 		{
-			psource = xf86CheckStrOption(pDevices->options, "_source", "");
-			pCommon = ((WacomDevicePtr)pDevices->private)->common;
+			WacomCommonPtr pCommon = ((WacomDevicePtr)pDevices->private)->common;
+			char* fsource = xf86CheckStrOption(pInfo->options, "_source", "");
+			char* psource = xf86CheckStrOption(pDevices->options, "_source", "");
+
 			if (pCommon->min_maj &&
 				pCommon->min_maj == min_maj)
 			{
@@ -78,14 +76,14 @@ static Bool wcmCheckSource(InputInfoPtr pInfo, dev_t min_maj)
  * the xorg.conf and is then hotplugged through the server backend (HAL,
  * udev). In this case, the hotplugged one fails.
  */
-int wcmIsDuplicate(char* device, InputInfoPtr pInfo)
+int wcmIsDuplicate(const char* device, InputInfoPtr pInfo)
 {
 	struct stat st;
 	int isInUse = 0;
-	char* lsource = xf86CheckStrOption(pInfo->options, "_source", "");
+	char* lsource = xf86CheckStrOption(pInfo->options, "_source", NULL);
 
 	/* always allow xorg.conf defined tools to be added */
-	if (!strlen(lsource)) goto ret;
+	if (!lsource || !strlen(lsource)) goto ret;
 
 	if (stat(device, &st) == -1)
 	{
@@ -136,7 +134,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 	int j, k, ret = FALSE;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
-	char* dsource = xf86CheckStrOption(pInfo->options, "_source", "");
+	char* dsource = xf86CheckStrOption(pInfo->options, "_source", NULL);
 
 	if (!type)
 	{
@@ -163,7 +161,7 @@ Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type)
 						    ret = FALSE;
 					}
 				}
-				else if (!strlen(dsource)) /* an user defined type */
+				else if (!dsource || !strlen(dsource)) /* an user defined type */
 				{
 					/* assume it is a valid type */
 					SETBIT(common->wcmKeys, wcmType[j].tool[k]);
@@ -191,8 +189,9 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 
 	switch (priv->common->tablet_id)
 	{
+		case 0xF8:  /* Cintiq 24HDT */
 		case 0xF4:  /* Cintiq 24HD */
-			TabletSetFeature(priv->common, WCM_DUALRING);
+			TabletSetFeature(priv->common, WCM_DUALRING | WCM_LCD);
 			/* fall through */
 
 		case 0x26:  /* I5 */
@@ -255,6 +254,7 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 		case 0xC5:  /* CintiqV5 */
 		case 0xC6:  /* CintiqV5 */
 		case 0xCC:  /* CinitqV5 */
+		case 0xFA:  /* Cintiq 22HD */
 			TabletSetFeature(priv->common, WCM_LCD);
 			/* fall through */
 		case 0xB0:  /* I3 */
@@ -267,20 +267,24 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 			TabletSetFeature(priv->common, WCM_STRIP | WCM_ROTATION);
 			break;
 
+		case 0x100: /* TPC with MT */
+		case 0x101: /* TPC with MT */
+		case 0x4001: /* TPC with MT */
 		case 0xE2: /* TPC with 2FGT */
 		case 0xE3: /* TPC with 2FGT */
+		case 0xE5: /* TPC with MT */
 		case 0xE6: /* TPC with 2FGT */
-			TabletSetFeature(priv->common, WCM_TPC);
-			break;
-
 		case 0x93: /* TPC with 1FGT */
 		case 0x9A: /* TPC with 1FGT */
+		case 0xED: /* TPC with 1FGT */
 		case 0x90: /* TPC */
 		case 0x97: /* TPC */
+		case 0xEF: /* TPC */
 			TabletSetFeature(priv->common, WCM_TPC);
 			break;
 
 		case 0x9F:
+		case 0xF6: /* Cintiq 24HDT Touch */
 			TabletSetFeature(priv->common, WCM_LCD);
 			break;
 	}
@@ -572,13 +576,16 @@ void wcmHotplugOthers(InputInfoPtr pInfo, const char *basename)
  */
 int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type)
 {
-	char *source = xf86CheckStrOption(pInfo->options, "_source", "");
+	char *source = xf86CheckStrOption(pInfo->options, "_source", NULL);
 	int i;
 
 	if (*type) /* type specified, don't hotplug */
 		return 0;
 
-	if (strcmp(source, "server/hal") && strcmp(source, "server/udev"))
+	if (!source) /* xorg.conf device, don't auto-pick type */
+		return 0;
+
+	if (source && strcmp(source, "server/hal") && strcmp(source, "server/udev"))
 		return 0;
 
 	/* no type specified, so we need to pick the first one applicable
@@ -706,7 +713,7 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 {
 	WacomDevicePtr  priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr  common = priv->common;
-	char            *s, b[12];
+	char            *s;
 	int		i;
 	WacomToolPtr    tool = NULL;
 	int		tpc_button_is_on;
@@ -736,8 +743,8 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 	 */
 	if (IsPad(priv))
 	{
-		priv->wheelup = priv->wheel2up = 4;
-		priv->wheeldn = priv->wheel2dn = 5;
+		priv->wheel_default[WHEEL_ABS_UP] = priv->wheel_default[WHEEL2_ABS_UP] = 4;
+		priv->wheel_default[WHEEL_ABS_DN] = priv->wheel_default[WHEEL2_ABS_DN] = 5;
 		set_absolute(pInfo, TRUE);
 	}
 
@@ -843,7 +850,7 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 
 		if(toollist) /* Already have a tool with the same type/serial */
 		{
-			xf86Msg(X_ERROR, "%s: already have a tool with type/serial %d/%d.",
+			xf86Msg(X_ERROR, "%s: already have a tool with type/serial %d/%d.\n",
 					pInfo->name, tool->typeid, tool->serial);
 			goto error;
 		} else /* No match on existing tool/serial, add tool to the end of the list */
@@ -889,6 +896,11 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 		else if (touch_is_on != common->wcmTouch)
 			xf86Msg(X_WARNING, "%s: Touch option can only be set "
 				"by a touch tool.\n", pInfo->name);
+
+		if (TabletHasFeature(common, WCM_1FGT))
+			common->wcmMaxContacts = 1;
+		else
+			common->wcmMaxContacts = 2;
 	}
 
 	/* 2FG touch device */
@@ -917,14 +929,15 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 	/* Swap stylus buttons 2 and 3 for Tablet PCs */
 	if (TabletHasFeature(common, WCM_TPC) && IsStylus(priv))
 	{
-		priv->button[1] = 3;
-		priv->button[2] = 2;
+		priv->button_default[1] = 3;
+		priv->button_default[2] = 2;
 	}
 
 	for (i=0; i<WCM_MAX_BUTTONS; i++)
 	{
+		char b[12];
 		sprintf(b, "Button%d", i+1);
-		priv->button[i] = xf86SetIntOption(pInfo->options, b, priv->button[i]);
+		priv->button_default[i] = xf86SetIntOption(pInfo->options, b, priv->button_default[i]);
 	}
 
 	/* Now parse class-specific options */
@@ -934,7 +947,6 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 
 	return TRUE;
 error:
-	free(tool);
 	return FALSE;
 }
 
