@@ -23,10 +23,9 @@
 #include <xorg-server.h>
 #include <xorgVersion.h>
 
+#include <wacom-util.h>
 #include "Xwacom.h"
 
-#define ARRAY_SIZE(a) (sizeof(a)/sizeof((a)[0]))
-
 /* max number of input events to read in one read call */
 #define MAX_EVENTS 50
 
@@ -38,6 +37,11 @@
 #include <xf86Xinput.h>
 #include <mipointer.h>
 #include <X11/Xatom.h>
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 18
+#define LogMessageVerbSigSafe xf86MsgVerb
+#endif
+
 /*****************************************************************************
  * Unit test hack
  ****************************************************************************/
@@ -57,9 +61,9 @@
 #define DBG(lvl, priv, ...) \
 	do { \
 		if ((lvl) <= priv->debugLevel) { \
-			xf86Msg(X_INFO, "%s (%d:%s): ", \
+			LogMessageVerbSigSafe(X_INFO, -1, "%s (%d:%s): ", \
 				((WacomDeviceRec*)priv)->name, lvl, __func__); \
-			xf86Msg(X_NONE, __VA_ARGS__); \
+			LogMessageVerbSigSafe(X_NONE, -1, __VA_ARGS__); \
 		} \
 	} while (0)
 #else
@@ -108,7 +112,7 @@ struct _WacomModule
 extern Bool wcmOpen(InputInfoPtr pInfo);
 
 /* device autoprobing */
-char *wcmEventAutoDevProbe (InputInfoPtr pInfo);
+const char *wcmEventAutoDevProbe (InputInfoPtr pInfo);
 
 /* common tablet initialization regime */
 int wcmInitTablet(InputInfoPtr pInfo, const char* id, float version);
@@ -132,7 +136,7 @@ void wcmMappingFactor(InputInfoPtr pInfo);
 /* validation */
 extern Bool wcmIsAValidType(InputInfoPtr pInfo, const char* type);
 extern Bool wcmIsWacomDevice (char* fname);
-extern int wcmIsDuplicate(char* device, InputInfoPtr pInfo);
+extern int wcmIsDuplicate(const char* device, InputInfoPtr pInfo);
 extern int wcmDeviceTypeKeys(InputInfoPtr pInfo);
 
 /* hotplug */
@@ -140,18 +144,18 @@ extern int wcmNeedAutoHotplug(InputInfoPtr pInfo, const char **type);
 extern void wcmHotplugOthers(InputInfoPtr pInfo, const char *basename);
 
 /* setup */
-extern int wcmParseOptions(InputInfoPtr pInfo, int hotplugged);
+extern Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent);
+extern Bool wcmPostInitParseOptions(InputInfoPtr pInfo, Bool is_primary, Bool is_dependent);
+extern int wcmParseSerials(InputInfoPtr pinfo);
 extern void wcmInitialCoordinates(InputInfoPtr pInfo, int axes);
 extern void wcmInitialScreens(InputInfoPtr pInfo);
-extern void wcmInitialScreens(InputInfoPtr pInfo);
 
 extern int wcmDevSwitchModeCall(InputInfoPtr pInfo, int mode);
 extern int wcmDevSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode);
 
 /* run-time modifications */
 extern void wcmChangeScreen(InputInfoPtr pInfo, int value);
-extern void wcmTilt2R(WacomDeviceStatePtr ds);
-extern void wcmGestureFilter(WacomDevicePtr priv, int channel);
+extern int wcmTilt2R(int x, int y, double offset);
 extern void wcmEmitKeycode(DeviceIntPtr keydev, int keycode, int state);
 extern void wcmSoftOutEvent(InputInfoPtr pInfo);
 
@@ -165,8 +169,11 @@ extern int wcmGetPhyDeviceID(WacomDevicePtr priv);
 
 /* device properties */
 extern int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop, BOOL checkonly);
+extern int wcmGetProperty(DeviceIntPtr dev, Atom property);
 extern int wcmDeleteProperty(DeviceIntPtr dev, Atom property);
 extern void InitWcmDeviceProperties(InputInfoPtr pInfo);
+extern void wcmUpdateRotationProperty(WacomDevicePtr priv);
+extern void wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial, int id);
 
 /* Utility functions */
 extern Bool is_absolute(InputInfoPtr pInfo);
@@ -175,6 +182,12 @@ extern WacomCommonPtr wcmRefCommon(WacomCommonPtr common);
 extern void wcmFreeCommon(WacomCommonPtr *common);
 extern WacomCommonPtr wcmNewCommon(void);
 
+enum WacomSuppressMode {
+	SUPPRESS_NONE = 8,	/* Process event normally */
+	SUPPRESS_ALL,		/* Supress and discard the whole event */
+	SUPPRESS_NON_MOTION	/* Supress all events but x/y motion */
+};
+
 /****************************************************************************/
 #endif /* __XF86WACOM_H */
 
